/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@pd.infn.it                            |
 | WWW:    http://www.pd.infn.it/~loreti/mlo.html       |
 *------------------------------------------------------*

 A small utility program for the AVL balanced binary tree
 routines; reads one or mode text files, counting words
 and reporting them in alphabetical order.

 $Id: countAvl.c,v 1.2 2002/12/17 06:42:19 loreti Exp $

 *------------------------------------------------------*/

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "avl.h"

#define WORD_LENGTH 128

static void insertItem(const char *, const unsigned int, const unsigned int,
                       const char *);
static int  compare(const item *, const item *);
static void delete(item *);
static void print(const item *);

node         *root   = 0;
unsigned int  nWords = 0;
unsigned int  nDiffs = 0;

int main(
  int   argc,
  char *argv[]
) {
  if (argc < 2) {
    fputs("Usage: ./countAvl <one or more file names>\n", stderr);
  } else {
    while (--argc) {
      FILE *fp;

      if ((fp = fopen(*++argv, "r")) == 0) {
        fprintf(stderr, "Error opening file \"%s\"\n", *argv);
        perror("fopen");
      } else {
        char         buffer[WORD_LENGTH];
        unsigned int line = 1;
        unsigned int next = 0;
        int          c;

        printf("Reading file \"%s\" ...", *argv);

        while ((c = getc(fp)) != EOF) {
          c = tolower((unsigned char) c);

          if (next == 0) {
            if (isalpha((unsigned char) c)) buffer[next++] = c;
          } else {
            if (isalpha((unsigned char) c)   ||
                isdigit((unsigned char) c)   ||   c == '_') {
              buffer[next++] = c;
            } else {
              buffer[next++] = '\0';
              insertItem(buffer, next, line, *argv);
              next = 0;
              if (c == '\n') line++;
            }
          }
        }

        if (ferror(fp)) {
          fprintf(stderr, "\nError reading from file \"%s\"\n", *argv);
          perror("fgets");
        } else {
          puts(" done.");
        }
        fclose(fp);

        if (next) {
          buffer[next++] = '\0';
          insertItem(buffer, next, line, *argv);
        }
      }
    }
  }

  fprintf(stderr, "\n%u words seen (%u different)\n\n",
          nWords, nDiffs);
  traverseAvl(root, INORDER, print);
  removeAvl(root, delete);
  return 0;
}

static void insertItem(
  const char         *word,
  const unsigned int  wordLen,
  const unsigned int  line,
  const char         *file
) {
  node *pNode;
  item *pItem = malloc(sizeof(item));
  assert(pItem != 0);

  pItem->lineNo = line;

  pItem->word = malloc(wordLen);
  assert(pItem->word != 0);
  strcpy(pItem->word, word);

  pItem->fileName = malloc(strlen(file)+1);
  assert(pItem->fileName != 0);
  strcpy(pItem->fileName, file);

  pItem->nOccur = 0;

  ++nWords;
  if (insertAvl(&root, pItem, &pNode, compare) == ERROR) {
    delete(pItem);
  } else {
    ++nDiffs;
  }

  ++(pNode->pItem->nOccur);
}

static int compare(
  const item *p1,
  const item *p2
) {
  return strcmp(p1->word, p2->word);
}

static void delete(
  item *p
) {
  free(p->word);
  free(p->fileName);
  free(p);
}

static void print(
  const item *p
) {
  printf("%s (first seen in file \"%s\", line %u): %u times\n",
         p->word, p->fileName, p->lineNo, p->nOccur);
}
