//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// Binary tree template class.
//
// Define the preprocessor symbol C_DEBUG in order to print to the
// error stream details about every call to a ctor or a dtor.
//
// $Id: bt.hh,v 1.1 2001/01/17 11:20:15 loreti Exp $
//
// -------------------------------------------------------------------

#ifndef BT_HH_
#define BT_HH_

#include <deque>

#ifdef C_DEBUG
#include <iostream>
using std::cerr;
using std::endl;
#endif // C_DEBUG

// Forward declarations

template <typename T> class bTree;
template <typename T> class treeSequencer;

// The template class btNode<T> represents a "node" in a binary tree
// containing data of type "T": a node has a data field holding its T
// value, plus two links to the children nodes.
//
// The copy constructor and the assignment operator of a btNode<T>
// (with a few other procedures) are declared private, in order to
// forbid to the user to call them directly; the user is supposed to
// use the methods of the class "bTree<T>", that abstracts the whole
// binary tree.
//
// The user should also supply two convenient methods "operator <" and
// "operator ==" to compare two T's, if they are not POD's.
//
// The methods supplied for a btNode<T> are:
// - a default constructor, and two constructors that may called with
//   a reference or a pointer to the data field (public);
// - a (virtual) destructor (public);
// - inspector and accessor functions, to get (or set) the left and
//   the right child (protected); and an inspector to get the data
//   field (public --- note that a value is assumed to be REALLY
//   STORED in the data field of the node, otherwise that method tries
//   to dereference a NULL pointer);
// - auxiliary procedures internally used to duplicate a node with its
//   subtree, and to remove a node with all its subtree (private);
// - copy constructor and assignment operator (private);
// - procedures to insert a given value and to retrieve a given value
//   in the subtree under a given node (private);
// - a procedure that modifies the links of the subtree under a given
//   node that will be later removed (private);
// - procedures to find the leftmost and the rightmost node in a given
//   subtree (private).

template <typename T> class btNode {
  friend class bTree<T>;
  friend class treeSequencer<T>;

private:
  T      * data_;
  btNode * left_;
  btNode * right_;

protected:
  virtual btNode * getLeft() const      { return left_; }
  virtual btNode * getRight() const     { return right_; }
  virtual void     setLeft(btNode * l)  { left_  = l; }
  virtual void     setRight(btNode * r) { right_ = r; }

  virtual void clone(const btNode & rhs) {
    setLeft(  rhs.getLeft()  ? new btNode( *(rhs.getLeft()) )  : 0 );
    setRight( rhs.getRight() ? new btNode( *(rhs.getRight()) ) : 0 );
    data_ = rhs.data_ ? new T( rhs.getData() ) : 0;
  }

  virtual void cleanup() {
    delete left_;
    delete right_;
    delete data_;
    left_ = right_ = 0;
    data_ = 0;
  }

#ifdef C_DEBUG
  virtual void printData() const {
    if (data_) cerr << getData(); else cerr << "<nil>";
  }

  void printLinks() const {
    cerr << " (left ";
    if (getLeft()) getLeft()->printData(); else cerr << "0";
    cerr << ", right ";
    if (getRight()) getRight()->printData(); else cerr << "0";
    cerr << ")\n";
  }

  void printAll() const { printData();  printLinks(); }
#endif // C_DEBUG

private:
  btNode(const btNode & rhs) {
#ifdef C_DEBUG
    cerr << "btNode copy ctor for ";
    rhs.printAll();
#endif // C_DEBUG
    clone(rhs);
  }

  const btNode & operator = (const btNode & rhs) {
    if (this != &rhs) {
#ifdef C_DEBUG
      cerr << "btNode assignment operator for ";
      rhs.printAll();
#endif // C_DEBUG
      cleanup();
      clone(rhs);
    }
    return *this;
  }

  virtual btNode * insert(const T & t, bool * newNode = 0);

  virtual btNode * find (const T & t) const;

  virtual btNode * remove();

  const btNode * leftmost() const {
    const btNode *           pNode = this;
    while (pNode->getLeft()) pNode = pNode->getLeft();
    return                   pNode;
  } // leftmost()

  const btNode * rightmost() const {
    const btNode *            pNode = this;
    while (pNode->getRight()) pNode = pNode->getRight();
    return                    pNode;
  } // rightmost()

public:
  btNode() : data_(0), left_(0), right_(0) {
#ifdef C_DEBUG
    cerr << "Default btNode ctor\n";
#endif // C_DEBUG
  }

  btNode(const T & t) : left_(0), right_(0) {
    data_ = new T(t);
#ifdef C_DEBUG
    cerr << "Reference btNode ctor for " << t << endl;
#endif // C_DEBUG
  }

  btNode(const T * pT) : left_(0), right_(0) {
    data_ = new T( *pT );
#ifdef C_DEBUG
    cerr << "Pointer btNode ctor for " << *pT << endl;
#endif // C_DEBUG
  }

  virtual ~btNode() {
#ifdef C_DEBUG
    cerr << "btNode dtor for ";
    printAll();
#endif // C_DEBUG
    cleanup();
  }

  T & getData () const { return *data_; }
};

// A template class bTree<T>, representing a binary tree of type "T"
// data; essentially a pointer to the root node, with some interface
// methods, in particular:
// - default constructor, copy constructor, assignment operator and
//   destructor (public);
// - inspector and accessor to the root node (protected);
// - a procedure to insert a new node in the tree (public);
// - a procedure to find a node holding a given value in the tree
//   (public);
// - two procedures to retrieve the leftmost or the rightmost node of
//   the tree (public);
// - a procedure to remove a node with a given value from the tree
//   (public).

template <typename T> class bTree {
  friend class treeSequencer<T>;

public:
  typedef T      value_type;
  typedef size_t size_type;

protected:
  btNode<T> * root_;

  virtual btNode<T> * getRoot() const            { return root_;  }
  virtual void        setRoot(btNode<T> * pNode) { root_ = pNode; }

private:
  virtual void clone(const bTree & rhs) {
    setRoot( rhs.getRoot() ? new btNode<T>( *(rhs.getRoot()) ) : 0 );
  }

public:
  bTree() : root_(0) {
#ifdef C_DEBUG
    cerr << "bTree default ctor\n";
#endif // C_DEBUG
  }

  virtual ~bTree() {
#ifdef C_DEBUG
    cerr << "bTree dtor for ";
    if (getRoot()) getRoot()->printAll();
    else           cerr << "nil\n";
#endif // C_DEBUG
    delete root_;
  }

  bTree(const bTree & rhs) {
#ifdef C_DEBUG
    cerr << "bTree copy ctor for ";
    if (rhs.getRoot()) rhs.getRoot()->printAll();
    else               cerr << "nil\n";
#endif // C_DEBUG
    clone(rhs);
  }

  const bTree & operator = (const bTree & rhs) {
    if (this != &rhs) {
#ifdef C_DEBUG
      cerr << "bTree assignment operator for ";
      if (rhs.getRoot()) rhs.getRoot()->printAll();
      else               cerr << "nil\n";
#endif // C_DEBUG
      delete root_;
      clone(rhs);
    }
    return *this;
  }

  virtual btNode<T> * insert(const T & t, bool * newNode = 0) {
    if (getRoot()) return getRoot()->insert(t, newNode);
    else {
      setRoot( new btNode<T>(t) );
      if (newNode) *newNode = true;
      return getRoot();
    }
  } // insert()

  virtual btNode<T> * find(const T & t) const {
    if (getRoot()) return getRoot()->find(t);
    else           return 0;
  } // find()

  virtual const btNode<T> * leftmost() const {
    if (getRoot()) return getRoot()->leftmost();
    else           return 0;
  } // leftmost()

  virtual const btNode<T> * rightmost() const {
    if (getRoot()) return getRoot()->rightmost();
    else           return 0;
  } // rightmost()

  virtual bool remove(const T & t);
};

// A pure virtual class to derive iterator-like objects for a binary
// tree

template <typename T> class sequencer {
public:
  virtual           ~sequencer() {}
  virtual bool       operator !  () const = 0;
  virtual const T &  operator () () const = 0;
  virtual void       operator ++ ()       = 0;
  virtual void       operator ++ (int)    = 0;
};

// A class derived from the above one, with the methods for the tree
// traversal; in the constructor the tree is sampled, and pointers to
// its data fields (in the correct order) are inserted in a double-end
// queue.
//
// The traversal order for every node may be: or left subtree, then
// the node, then the right subtree ("in-order", the default); or
// node, left subtree, right subtree ("pre-order"); or left subtree,
// right subtree, node ("postorder").  The user may optionally select
// to walk the tree in the reverse order, for any of these three
// sequences.
//
// Methods are supplied (all public):
// - to initialize the data sequence from a given binary tree (ctor);
// - to test for the end of the data (operator !);
// - to retrieve the current data (operator ());
// - to advance to the next data (prefix or infix operator ++).
//
// The direct use of copy constructor and assignment operator is
// forbidden.


template <typename T> class treeSequencer : public sequencer<T> {
public:
  enum seqType { INORDER, PREORDER, POSTORDER };

private:
  treeSequencer(const treeSequencer &);
  const treeSequencer & operator = (const treeSequencer &);

protected:
  seqType               sequence_;
  bool                  reverse_;
  std::deque<const T *> q_;

  void pushNode(const btNode<T> * p) {
    if (sequence_ == PREORDER)  q_.push_back(p->data_);
    if (p->getLeft())           pushNode(p->getLeft());
    if (sequence_ == INORDER)   q_.push_back(p->data_);
    if (p->getRight())          pushNode(p->getRight());
    if (sequence_ == POSTORDER) q_.push_back(p->data_);
  }

public:
  treeSequencer(const bTree<T> & t, seqType s = INORDER,
                bool r = false) : sequence_(s), reverse_(r)
    {
      if (t.getRoot()) pushNode(t.getRoot());
    }

  bool operator ! ()  const {
    return ! q_.empty();
  }

  const T & operator () () const {
    if (reverse_) return *( q_.back()  );
    else          return *( q_.front() ) ;
  }

  void operator ++ () {
    if (reverse_) q_.pop_back();
    else          q_.pop_front();
  }

  void operator ++ (int i) {
    if (reverse_) q_.pop_back();
    else          q_.pop_front();
  }
};

// From here on, non-inline methods that should go in a file bt.cxx;
// but g++ fails to instantiate all the templates if they are not
// here.

template <typename T>
btNode<T> * btNode<T>::insert(
  const T & t,
  bool    * newNode
) {

  // Tries to insert a node with the supplied data field in the
  // subtree lying under a given node, if it does not exist already.
  // Returns a pointer to the node holding the wanted value; and (if
  // "newNode" is not NULL) stores in *newNode 'true' if a new node
  // has been created, or 'false' if a node holding the value "t" was
  // already present.

  if (t < getData()) {
    if (getLeft()) {
      return getLeft()->insert(t, newNode);
    } else {
      setLeft( new btNode(t) );
      if (newNode) *newNode = true;
      return getLeft();
    }
  } else if (getData() < t) {
    if (getRight()) {
      return getRight()->insert(t, newNode);
    } else {
      setRight( new btNode(t) );
      if (newNode) *newNode = true;
      return getRight();
    }
  } else {
    if (newNode) *newNode = false;
    return this;
  }
} // insert()

template <typename T>
btNode<T> * btNode<T>::find (
  const T & t
) const {

  // Tries to find a node with a "t" data field in the subtree under
  // the given node.  Returns 0 if the node does not exist, or a
  // pointer to it.

  if (t < getData()) {
    if (getLeft()) return getLeft()->find(t);
  } else if (getData() < t) {
    if (getRight()) return getRight()->find(t);
  } else {
    return const_cast< btNode<T> * >(this);
  }
  return 0;
} // find()

template <typename T>
btNode<T> * btNode<T>::remove()
{

  // The current node has to be removed from the tree; remove() takes
  // care of the links in the node itself and the underlying subtree,
  // (but does not delete the node).

  // We test first for the special case of a node having only a single
  // child; then for the other special case of a right child without a
  // left subtree

  if (getLeft() == 0)  return getRight();
  if (getRight() == 0) return getLeft();

  btNode * child = getRight()->getLeft();
  if (child == 0) {
    getRight()->setLeft( getLeft() );
    return getRight();
  }

  // For the general case, we slide left in the right subtree; then
  // disconnect the leftmost node, to be moved to the top.

  btNode * parent = getRight();
  while (child->getLeft()) {
    parent = child;
    child  = child->getLeft();
  }

  parent->setLeft( child->getRight() );
  child->setLeft( getLeft() );
  child->setRight( getRight() );

  return child;
}

template <typename T>
bool bTree<T>::remove(
  const T & t
) {

  // Removes the node holding a given value from the binary tree: and
  // returns 'true' if successful, 'false' if the value is not in the
  // tree.
  // - looks for the parent of the node holding "t", noting if the
  //   node to be removed is the left or the right child;
  // - calls remove() for the node, to update the links of the parent
  //   and its subtree;
  // - deletes the node, after having set both its links to zero (to
  //   prevent the remotion of the subtrees).

  btNode<T> * pNode;

  // Special cases: (1) an empty tree; (2) the wanted node is the
  // root.

  if (getRoot() == 0) return false;

  if (t == getRoot()->getData()) {
    pNode = getRoot()->remove();

    getRoot()->setLeft(0);
    getRoot()->setRight(0);
    delete root_;

    setRoot( pNode );
    return true;
  }

  btNode<T> * next = 0;

  for (pNode = getRoot();   pNode;   pNode = next) {
    if (t < pNode->getData()) {
      next = pNode->getLeft();

      if (next   &&   next->getData() == t) {
        pNode->setLeft( next->remove() );

        next->setLeft(0);
        next->setRight(0);
        delete next;
        return true;
      }
    } else if (pNode->getData() < t) {
      next = pNode->getRight();

      if (next   &&   next->getData() == t) {
        pNode->setRight( next->remove() );

        next->setLeft(0);
        next->setRight(0);
        delete next;
        return true;
      }
    }
  }

  return false;
} // remove()

#endif // BT_HH_
